rem ************************************************************************
rem * uninstall-headers.bat - generated by gendsp.pl
rem *
set msvc=%1
del %COINDIR%\include\Inventor\C\XML\attribute.h
del %COINDIR%\include\Inventor\C\XML\document.h
del %COINDIR%\include\Inventor\C\XML\element.h
del %COINDIR%\include\Inventor\C\XML\entity.h
del %COINDIR%\include\Inventor\C\XML\path.h
del %COINDIR%\include\Inventor\C\XML\types.h
del %COINDIR%\include\Inventor\C\XML\parser.h
del %COINDIR%\include\Inventor\C\XML\world.h
del %COINDIR%\include\Inventor\C\base\hash.h
del %COINDIR%\include\Inventor\C\base\heap.h
del %COINDIR%\include\Inventor\C\base\memalloc.h
del %COINDIR%\include\Inventor\C\base\rbptree.h
del %COINDIR%\include\Inventor\C\base\time.h
del %COINDIR%\include\Inventor\C\base\string.h
del %COINDIR%\include\Inventor\C\base\list.h
del %COINDIR%\include\Inventor\C\errors\error.h
del %COINDIR%\include\Inventor\C\errors\debugerror.h
del %COINDIR%\include\Inventor\C\glue\gl.h
del %COINDIR%\include\Inventor\C\glue\dl.h
del %COINDIR%\include\Inventor\C\glue\spidermonkey.h
del %COINDIR%\include\Inventor\C\threads\common.h
del %COINDIR%\include\Inventor\C\threads\thread.h
del %COINDIR%\include\Inventor\C\threads\mutex.h
del %COINDIR%\include\Inventor\C\threads\condvar.h
del %COINDIR%\include\Inventor\C\threads\recmutex.h
del %COINDIR%\include\Inventor\C\threads\rwmutex.h
del %COINDIR%\include\Inventor\C\threads\storage.h
del %COINDIR%\include\Inventor\C\threads\worker.h
del %COINDIR%\include\Inventor\C\threads\wpool.h
del %COINDIR%\include\Inventor\C\threads\sched.h
del %COINDIR%\include\Inventor\C\threads\sync.h
del %COINDIR%\include\Inventor\C\threads\fifo.h
del %COINDIR%\include\Inventor\C\threads\barrier.h
del %COINDIR%\include\Inventor\C\basic.h
del %COINDIR%\include\Inventor\C\tidbits.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRML.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLAnchor.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLAppearance.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLAudioClip.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLBackground.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLBillboard.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLBox.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLCollision.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLColor.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLColorInterpolator.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLCone.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLCoordinate.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLCoordinateInterpolator.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLCylinder.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLCylinderSensor.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLDirectionalLight.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLDragSensor.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLElevationGrid.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLExtrusion.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLFog.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLFontStyle.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLGeometry.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLGroup.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLImageTexture.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLIndexedFaceSet.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLIndexedLine.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLIndexedLineSet.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLIndexedShape.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLInline.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLInterpolator.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLLOD.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLLight.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLMacros.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLMaterial.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLMovieTexture.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLNavigationInfo.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLNodes.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLNormal.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLNormalInterpolator.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLOrientationInterpolator.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLParent.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLPixelTexture.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLPlaneSensor.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLPointLight.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLPointSet.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLPositionInterpolator.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLProximitySensor.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLScalarInterpolator.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLScript.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLSensor.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLShape.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLSound.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLSphere.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLSphereSensor.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLSpotLight.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLSubInterpolator.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLSwitch.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLText.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLTexture.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLTextureCoordinate.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLTextureTransform.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLTimeSensor.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLTouchSensor.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLTransform.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLVertexLine.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLVertexPoint.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLVertexShape.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLViewpoint.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLVisibilitySensor.h
del %COINDIR%\include\Inventor\VRMLnodes\SoVRMLWorldInfo.h
del %COINDIR%\include\Inventor\actions\SoSubAction.h
del %COINDIR%\include\Inventor\actions\SoActions.h
del %COINDIR%\include\Inventor\actions\SoAction.h
del %COINDIR%\include\Inventor\actions\SoBoxHighlightRenderAction.h
del %COINDIR%\include\Inventor\actions\SoCallbackAction.h
del %COINDIR%\include\Inventor\actions\SoGLRenderAction.h
del %COINDIR%\include\Inventor\actions\SoGetBoundingBoxAction.h
del %COINDIR%\include\Inventor\actions\SoGetMatrixAction.h
del %COINDIR%\include\Inventor\actions\SoGetPrimitiveCountAction.h
del %COINDIR%\include\Inventor\actions\SoHandleEventAction.h
del %COINDIR%\include\Inventor\actions\SoLineHighlightRenderAction.h
del %COINDIR%\include\Inventor\actions\SoPickAction.h
del %COINDIR%\include\Inventor\actions\SoRayPickAction.h
del %COINDIR%\include\Inventor\actions\SoReorganizeAction.h
del %COINDIR%\include\Inventor\actions\SoSearchAction.h
del %COINDIR%\include\Inventor\actions\SoSimplifyAction.h
del %COINDIR%\include\Inventor\actions\SoToVRMLAction.h
del %COINDIR%\include\Inventor\actions\SoToVRML2Action.h
del %COINDIR%\include\Inventor\actions\SoWriteAction.h
del %COINDIR%\include\Inventor\actions\SoAudioRenderAction.h
del %COINDIR%\include\Inventor\annex\HardCopy\SoHardCopy.h
del %COINDIR%\include\Inventor\annex\HardCopy\SoPSVectorOutput.h
del %COINDIR%\include\Inventor\annex\HardCopy\SoVectorOutput.h
del %COINDIR%\include\Inventor\annex\HardCopy\SoVectorizeAction.h
del %COINDIR%\include\Inventor\annex\HardCopy\SoVectorizePSAction.h
del %COINDIR%\include\Inventor\annex\ForeignFiles\SoForeignFileKit.h
del %COINDIR%\include\Inventor\annex\ForeignFiles\SoSTLFileKit.h
del %COINDIR%\include\Inventor\annex\FXViz\nodes\SoShadowGroup.h
del %COINDIR%\include\Inventor\annex\FXViz\nodes\SoShadowStyle.h
del %COINDIR%\include\Inventor\annex\FXViz\nodes\SoShadowDirectionalLight.h
del %COINDIR%\include\Inventor\annex\FXViz\nodes\SoShadowSpotLight.h
del %COINDIR%\include\Inventor\annex\FXViz\nodes\SoShadowCulling.h
del %COINDIR%\include\Inventor\annex\FXViz\elements\SoShadowStyleElement.h
del %COINDIR%\include\Inventor\annex\FXViz\elements\SoGLShadowCullingElement.h
del %COINDIR%\include\Inventor\annex\Profiler\nodes\SoProfilerStats.h
del %COINDIR%\include\Inventor\annex\Profiler\elements\SoProfilerElement.h
del %COINDIR%\include\Inventor\annex\Profiler\nodekits\SoNodeVisualize.h
del %COINDIR%\include\Inventor\annex\Profiler\nodekits\SoProfilerOverlayKit.h
del %COINDIR%\include\Inventor\annex\Profiler\nodekits\SoProfilerTopKit.h
del %COINDIR%\include\Inventor\annex\Profiler\nodekits\SoScrollingGraphKit.h
del %COINDIR%\include\Inventor\annex\Profiler\nodekits\SoProfilerVisualizeKit.h
del %COINDIR%\include\Inventor\annex\Profiler\engines\SoProfilerTopEngine.h
del %COINDIR%\include\Inventor\annex\Profiler\utils\SoProfilingReportGenerator.h
del %COINDIR%\include\Inventor\annex\Profiler\SbProfilingData.h
del %COINDIR%\include\Inventor\annex\Profiler\SoProfiler.h
del %COINDIR%\include\Inventor\bundles\SoBundle.h
del %COINDIR%\include\Inventor\bundles\SoMaterialBundle.h
del %COINDIR%\include\Inventor\bundles\SoNormalBundle.h
del %COINDIR%\include\Inventor\bundles\SoVertexAttributeBundle.h
del %COINDIR%\include\Inventor\bundles\SoTextureCoordinateBundle.h
del %COINDIR%\include\Inventor\caches\SoBoundingBoxCache.h
del %COINDIR%\include\Inventor\caches\SoCache.h
del %COINDIR%\include\Inventor\caches\SoConvexDataCache.h
del %COINDIR%\include\Inventor\caches\SoGLCacheList.h
del %COINDIR%\include\Inventor\caches\SoGLRenderCache.h
del %COINDIR%\include\Inventor\caches\SoNormalCache.h
del %COINDIR%\include\Inventor\caches\SoPrimitiveVertexCache.h
del %COINDIR%\include\Inventor\caches\SoTextureCoordinateCache.h
del %COINDIR%\include\Inventor\collision\SoIntersectionDetectionAction.h
del %COINDIR%\include\Inventor\details\SoSubDetail.h
del %COINDIR%\include\Inventor\details\SoDetail.h
del %COINDIR%\include\Inventor\details\SoDetails.h
del %COINDIR%\include\Inventor\details\SoConeDetail.h
del %COINDIR%\include\Inventor\details\SoCubeDetail.h
del %COINDIR%\include\Inventor\details\SoCylinderDetail.h
del %COINDIR%\include\Inventor\details\SoFaceDetail.h
del %COINDIR%\include\Inventor\details\SoLineDetail.h
del %COINDIR%\include\Inventor\details\SoNodeKitDetail.h
del %COINDIR%\include\Inventor\details\SoPointDetail.h
del %COINDIR%\include\Inventor\details\SoTextDetail.h
del %COINDIR%\include\Inventor\draggers\SoDragger.h
del %COINDIR%\include\Inventor\draggers\SoCenterballDragger.h
del %COINDIR%\include\Inventor\draggers\SoDirectionalLightDragger.h
del %COINDIR%\include\Inventor\draggers\SoDragPointDragger.h
del %COINDIR%\include\Inventor\draggers\SoHandleBoxDragger.h
del %COINDIR%\include\Inventor\draggers\SoJackDragger.h
del %COINDIR%\include\Inventor\draggers\SoPointLightDragger.h
del %COINDIR%\include\Inventor\draggers\SoRotateCylindricalDragger.h
del %COINDIR%\include\Inventor\draggers\SoRotateDiscDragger.h
del %COINDIR%\include\Inventor\draggers\SoRotateSphericalDragger.h
del %COINDIR%\include\Inventor\draggers\SoScale1Dragger.h
del %COINDIR%\include\Inventor\draggers\SoScale2Dragger.h
del %COINDIR%\include\Inventor\draggers\SoScale2UniformDragger.h
del %COINDIR%\include\Inventor\draggers\SoScaleUniformDragger.h
del %COINDIR%\include\Inventor\draggers\SoSpotLightDragger.h
del %COINDIR%\include\Inventor\draggers\SoTabBoxDragger.h
del %COINDIR%\include\Inventor\draggers\SoTabPlaneDragger.h
del %COINDIR%\include\Inventor\draggers\SoTrackballDragger.h
del %COINDIR%\include\Inventor\draggers\SoTransformBoxDragger.h
del %COINDIR%\include\Inventor\draggers\SoTransformerDragger.h
del %COINDIR%\include\Inventor\draggers\SoTranslate1Dragger.h
del %COINDIR%\include\Inventor\draggers\SoTranslate2Dragger.h
del %COINDIR%\include\Inventor\elements\SoSubElement.h
del %COINDIR%\include\Inventor\elements\SoElements.h
del %COINDIR%\include\Inventor\elements\SoAccumulatedElement.h
del %COINDIR%\include\Inventor\elements\SoAmbientColorElement.h
del %COINDIR%\include\Inventor\elements\SoAnnoText3CharOrientElement.h
del %COINDIR%\include\Inventor\elements\SoAnnoText3FontSizeHintElement.h
del %COINDIR%\include\Inventor\elements\SoAnnoText3RenderPrintElement.h
del %COINDIR%\include\Inventor\elements\SoBBoxModelMatrixElement.h
del %COINDIR%\include\Inventor\elements\SoBumpMapElement.h
del %COINDIR%\include\Inventor\elements\SoBumpMapCoordinateElement.h
del %COINDIR%\include\Inventor\elements\SoBumpMapMatrixElement.h
del %COINDIR%\include\Inventor\elements\SoCacheElement.h
del %COINDIR%\include\Inventor\elements\SoCacheHintElement.h
del %COINDIR%\include\Inventor\elements\SoClipPlaneElement.h
del %COINDIR%\include\Inventor\elements\SoComplexityElement.h
del %COINDIR%\include\Inventor\elements\SoComplexityTypeElement.h
del %COINDIR%\include\Inventor\elements\SoCoordinateElement.h
del %COINDIR%\include\Inventor\elements\SoCreaseAngleElement.h
del %COINDIR%\include\Inventor\elements\SoCullElement.h
del %COINDIR%\include\Inventor\elements\SoDecimationPercentageElement.h
del %COINDIR%\include\Inventor\elements\SoDecimationTypeElement.h
del %COINDIR%\include\Inventor\elements\SoDiffuseColorElement.h
del %COINDIR%\include\Inventor\elements\SoDrawStyleElement.h
del %COINDIR%\include\Inventor\elements\SoElement.h
del %COINDIR%\include\Inventor\elements\SoEmissiveColorElement.h
del %COINDIR%\include\Inventor\elements\SoEnvironmentElement.h
del %COINDIR%\include\Inventor\elements\SoFloatElement.h
del %COINDIR%\include\Inventor\elements\SoFocalDistanceElement.h
del %COINDIR%\include\Inventor\elements\SoFontNameElement.h
del %COINDIR%\include\Inventor\elements\SoFontSizeElement.h
del %COINDIR%\include\Inventor\elements\SoGeoElement.h
del %COINDIR%\include\Inventor\elements\SoGLCacheContextElement.h
del %COINDIR%\include\Inventor\elements\SoGLClipPlaneElement.h
del %COINDIR%\include\Inventor\elements\SoGLColorIndexElement.h
del %COINDIR%\include\Inventor\elements\SoGLCoordinateElement.h
del %COINDIR%\include\Inventor\elements\SoGLDisplayList.h
del %COINDIR%\include\Inventor\elements\SoGLDrawStyleElement.h
del %COINDIR%\include\Inventor\elements\SoGLEnvironmentElement.h
del %COINDIR%\include\Inventor\elements\SoGLLazyElement.h
del %COINDIR%\include\Inventor\elements\SoGLLightIdElement.h
del %COINDIR%\include\Inventor\elements\SoGLLinePatternElement.h
del %COINDIR%\include\Inventor\elements\SoGLLineWidthElement.h
del %COINDIR%\include\Inventor\elements\SoGLModelMatrixElement.h
del %COINDIR%\include\Inventor\elements\SoGLMultiTextureCoordinateElement.h
del %COINDIR%\include\Inventor\elements\SoGLMultiTextureEnabledElement.h
del %COINDIR%\include\Inventor\elements\SoGLMultiTextureMatrixElement.h
del %COINDIR%\include\Inventor\elements\SoGLMultiTextureImageElement.h
del %COINDIR%\include\Inventor\elements\SoGLNormalElement.h
del %COINDIR%\include\Inventor\elements\SoGLNormalizeElement.h
del %COINDIR%\include\Inventor\elements\SoGLPointSizeElement.h
del %COINDIR%\include\Inventor\elements\SoGLPolygonOffsetElement.h
del %COINDIR%\include\Inventor\elements\SoGLProjectionMatrixElement.h
del %COINDIR%\include\Inventor\elements\SoGLRenderPassElement.h
del %COINDIR%\include\Inventor\elements\SoGLShapeHintsElement.h
del %COINDIR%\include\Inventor\elements\SoGLShadeModelElement.h
del %COINDIR%\include\Inventor\elements\SoGLTextureCoordinateElement.h
del %COINDIR%\include\Inventor\elements\SoGLTextureEnabledElement.h
del %COINDIR%\include\Inventor\elements\SoGLTexture3EnabledElement.h
del %COINDIR%\include\Inventor\elements\SoGLTextureImageElement.h
del %COINDIR%\include\Inventor\elements\SoGLTextureMatrixElement.h
del %COINDIR%\include\Inventor\elements\SoGLUpdateAreaElement.h
del %COINDIR%\include\Inventor\elements\SoGLVBOElement.h
del %COINDIR%\include\Inventor\elements\SoGLViewingMatrixElement.h
del %COINDIR%\include\Inventor\elements\SoGLViewportRegionElement.h
del %COINDIR%\include\Inventor\elements\SoInt32Element.h
del %COINDIR%\include\Inventor\elements\SoLazyElement.h
del %COINDIR%\include\Inventor\elements\SoLightAttenuationElement.h
del %COINDIR%\include\Inventor\elements\SoLightElement.h
del %COINDIR%\include\Inventor\elements\SoLightModelElement.h
del %COINDIR%\include\Inventor\elements\SoLinePatternElement.h
del %COINDIR%\include\Inventor\elements\SoLineWidthElement.h
del %COINDIR%\include\Inventor\elements\SoLocalBBoxMatrixElement.h
del %COINDIR%\include\Inventor\elements\SoLongElement.h
del %COINDIR%\include\Inventor\elements\SoMaterialBindingElement.h
del %COINDIR%\include\Inventor\elements\SoVertexAttributeBindingElement.h
del %COINDIR%\include\Inventor\elements\SoModelMatrixElement.h
del %COINDIR%\include\Inventor\elements\SoMultiTextureCoordinateElement.h
del %COINDIR%\include\Inventor\elements\SoMultiTextureEnabledElement.h
del %COINDIR%\include\Inventor\elements\SoMultiTextureMatrixElement.h
del %COINDIR%\include\Inventor\elements\SoMultiTextureImageElement.h
del %COINDIR%\include\Inventor\elements\SoNormalBindingElement.h
del %COINDIR%\include\Inventor\elements\SoNormalElement.h
del %COINDIR%\include\Inventor\elements\SoOverrideElement.h
del %COINDIR%\include\Inventor\elements\SoPickRayElement.h
del %COINDIR%\include\Inventor\elements\SoPickStyleElement.h
del %COINDIR%\include\Inventor\elements\SoPointSizeElement.h
del %COINDIR%\include\Inventor\elements\SoPolygonOffsetElement.h
del %COINDIR%\include\Inventor\elements\SoProfileCoordinateElement.h
del %COINDIR%\include\Inventor\elements\SoProfileElement.h
del %COINDIR%\include\Inventor\elements\SoProjectionMatrixElement.h
del %COINDIR%\include\Inventor\elements\SoReplacedElement.h
del %COINDIR%\include\Inventor\elements\SoShapeHintsElement.h
del %COINDIR%\include\Inventor\elements\SoShapeStyleElement.h
del %COINDIR%\include\Inventor\elements\SoShininessElement.h
del %COINDIR%\include\Inventor\elements\SoSpecularColorElement.h
del %COINDIR%\include\Inventor\elements\SoSwitchElement.h
del %COINDIR%\include\Inventor\elements\SoTextOutlineEnabledElement.h
del %COINDIR%\include\Inventor\elements\SoTextureCombineElement.h
del %COINDIR%\include\Inventor\elements\SoTextureCoordinateBindingElement.h
del %COINDIR%\include\Inventor\elements\SoTextureCoordinateElement.h
del %COINDIR%\include\Inventor\elements\SoTextureEnabledElement.h
del %COINDIR%\include\Inventor\elements\SoTexture3EnabledElement.h
del %COINDIR%\include\Inventor\elements\SoTextureImageElement.h
del %COINDIR%\include\Inventor\elements\SoTextureMatrixElement.h
del %COINDIR%\include\Inventor\elements\SoTextureOverrideElement.h
del %COINDIR%\include\Inventor\elements\SoTextureQualityElement.h
del %COINDIR%\include\Inventor\elements\SoTextureUnitElement.h
del %COINDIR%\include\Inventor\elements\SoTransparencyElement.h
del %COINDIR%\include\Inventor\elements\SoUnitsElement.h
del %COINDIR%\include\Inventor\elements\SoViewVolumeElement.h
del %COINDIR%\include\Inventor\elements\SoViewingMatrixElement.h
del %COINDIR%\include\Inventor\elements\SoViewportRegionElement.h
del %COINDIR%\include\Inventor\elements\SoWindowElement.h
del %COINDIR%\include\Inventor\elements\SoListenerPositionElement.h
del %COINDIR%\include\Inventor\elements\SoListenerOrientationElement.h
del %COINDIR%\include\Inventor\elements\SoListenerGainElement.h
del %COINDIR%\include\Inventor\elements\SoListenerDopplerElement.h
del %COINDIR%\include\Inventor\elements\SoSoundElement.h
del %COINDIR%\include\Inventor\elements\SoGLShaderProgramElement.h
del %COINDIR%\include\Inventor\elements\SoDepthBufferElement.h
del %COINDIR%\include\Inventor\elements\SoGLDepthBufferElement.h
del %COINDIR%\include\Inventor\elements\SoVertexAttributeElement.h
del %COINDIR%\include\Inventor\elements\SoGLVertexAttributeElement.h
del %COINDIR%\include\Inventor\engines\SoSubEngine.h
del %COINDIR%\include\Inventor\engines\SoSubNodeEngine.h
del %COINDIR%\include\Inventor\engines\SoEngines.h
del %COINDIR%\include\Inventor\engines\SoBoolOperation.h
del %COINDIR%\include\Inventor\engines\SoCalculator.h
del %COINDIR%\include\Inventor\engines\SoCompose.h
del %COINDIR%\include\Inventor\engines\SoComposeMatrix.h
del %COINDIR%\include\Inventor\engines\SoComposeRotation.h
del %COINDIR%\include\Inventor\engines\SoComposeRotationFromTo.h
del %COINDIR%\include\Inventor\engines\SoComposeVec2f.h
del %COINDIR%\include\Inventor\engines\SoComposeVec3f.h
del %COINDIR%\include\Inventor\engines\SoComposeVec4f.h
del %COINDIR%\include\Inventor\engines\SoComputeBoundingBox.h
del %COINDIR%\include\Inventor\engines\SoConcatenate.h
del %COINDIR%\include\Inventor\engines\SoCounter.h
del %COINDIR%\include\Inventor\engines\SoDecomposeMatrix.h
del %COINDIR%\include\Inventor\engines\SoDecomposeRotation.h
del %COINDIR%\include\Inventor\engines\SoDecomposeVec2f.h
del %COINDIR%\include\Inventor\engines\SoDecomposeVec3f.h
del %COINDIR%\include\Inventor\engines\SoDecomposeVec4f.h
del %COINDIR%\include\Inventor\engines\SoElapsedTime.h
del %COINDIR%\include\Inventor\engines\SoEngine.h
del %COINDIR%\include\Inventor\engines\SoEngineOutput.h
del %COINDIR%\include\Inventor\engines\SoFieldConverter.h
del %COINDIR%\include\Inventor\engines\SoGate.h
del %COINDIR%\include\Inventor\engines\SoInterpolate.h
del %COINDIR%\include\Inventor\engines\SoInterpolateFloat.h
del %COINDIR%\include\Inventor\engines\SoInterpolateRotation.h
del %COINDIR%\include\Inventor\engines\SoInterpolateVec2f.h
del %COINDIR%\include\Inventor\engines\SoInterpolateVec3f.h
del %COINDIR%\include\Inventor\engines\SoInterpolateVec4f.h
del %COINDIR%\include\Inventor\engines\SoNodeEngine.h
del %COINDIR%\include\Inventor\engines\SoOnOff.h
del %COINDIR%\include\Inventor\engines\SoOneShot.h
del %COINDIR%\include\Inventor\engines\SoOutputData.h
del %COINDIR%\include\Inventor\engines\SoSelectOne.h
del %COINDIR%\include\Inventor\engines\SoTimeCounter.h
del %COINDIR%\include\Inventor\engines\SoTransformVec3f.h
del %COINDIR%\include\Inventor\engines\SoTriggerAny.h
del %COINDIR%\include\Inventor\engines\SoTexture2Convert.h
del %COINDIR%\include\Inventor\engines\SoHeightMapToNormalMap.h
del %COINDIR%\include\Inventor\errors\SoErrors.h
del %COINDIR%\include\Inventor\errors\SoDebugError.h
del %COINDIR%\include\Inventor\errors\SoError.h
del %COINDIR%\include\Inventor\errors\SoMemoryError.h
del %COINDIR%\include\Inventor\errors\SoReadError.h
del %COINDIR%\include\Inventor\events\SoSubEvent.h
del %COINDIR%\include\Inventor\events\SoButtonEvent.h
del %COINDIR%\include\Inventor\events\SoEvent.h
del %COINDIR%\include\Inventor\events\SoEvents.h
del %COINDIR%\include\Inventor\events\SoKeyboardEvent.h
del %COINDIR%\include\Inventor\events\SoLocation2Event.h
del %COINDIR%\include\Inventor\events\SoMotion3Event.h
del %COINDIR%\include\Inventor\events\SoMouseButtonEvent.h
del %COINDIR%\include\Inventor\events\SoSpaceballButtonEvent.h
del %COINDIR%\include\Inventor\fields\SoSubField.h
del %COINDIR%\include\Inventor\fields\SoFields.h
del %COINDIR%\include\Inventor\fields\SoField.h
del %COINDIR%\include\Inventor\fields\SoFieldContainer.h
del %COINDIR%\include\Inventor\fields\SoFieldData.h
del %COINDIR%\include\Inventor\fields\SoMFBitMask.h
del %COINDIR%\include\Inventor\fields\SoMFBool.h
del %COINDIR%\include\Inventor\fields\SoMFColor.h
del %COINDIR%\include\Inventor\fields\SoMFColorRGBA.h
del %COINDIR%\include\Inventor\fields\SoMFDouble.h
del %COINDIR%\include\Inventor\fields\SoMFEngine.h
del %COINDIR%\include\Inventor\fields\SoMFEnum.h
del %COINDIR%\include\Inventor\fields\SoMFFloat.h
del %COINDIR%\include\Inventor\fields\SoMFInt32.h
del %COINDIR%\include\Inventor\fields\SoMFLong.h
del %COINDIR%\include\Inventor\fields\SoMFMatrix.h
del %COINDIR%\include\Inventor\fields\SoMFName.h
del %COINDIR%\include\Inventor\fields\SoMFNode.h
del %COINDIR%\include\Inventor\fields\SoMFPath.h
del %COINDIR%\include\Inventor\fields\SoMFPlane.h
del %COINDIR%\include\Inventor\fields\SoMFRotation.h
del %COINDIR%\include\Inventor\fields\SoMFShort.h
del %COINDIR%\include\Inventor\fields\SoMFString.h
del %COINDIR%\include\Inventor\fields\SoMFTime.h
del %COINDIR%\include\Inventor\fields\SoMFUInt32.h
del %COINDIR%\include\Inventor\fields\SoMFULong.h
del %COINDIR%\include\Inventor\fields\SoMFUShort.h
del %COINDIR%\include\Inventor\fields\SoMFVec2b.h
del %COINDIR%\include\Inventor\fields\SoMFVec2s.h
del %COINDIR%\include\Inventor\fields\SoMFVec2i32.h
del %COINDIR%\include\Inventor\fields\SoMFVec2f.h
del %COINDIR%\include\Inventor\fields\SoMFVec2d.h
del %COINDIR%\include\Inventor\fields\SoMFVec3b.h
del %COINDIR%\include\Inventor\fields\SoMFVec3s.h
del %COINDIR%\include\Inventor\fields\SoMFVec3i32.h
del %COINDIR%\include\Inventor\fields\SoMFVec3f.h
del %COINDIR%\include\Inventor\fields\SoMFVec3d.h
del %COINDIR%\include\Inventor\fields\SoMFVec4b.h
del %COINDIR%\include\Inventor\fields\SoMFVec4ub.h
del %COINDIR%\include\Inventor\fields\SoMFVec4s.h
del %COINDIR%\include\Inventor\fields\SoMFVec4us.h
del %COINDIR%\include\Inventor\fields\SoMFVec4i32.h
del %COINDIR%\include\Inventor\fields\SoMFVec4ui32.h
del %COINDIR%\include\Inventor\fields\SoMFVec4f.h
del %COINDIR%\include\Inventor\fields\SoMFVec4d.h
del %COINDIR%\include\Inventor\fields\SoMField.h
del %COINDIR%\include\Inventor\fields\SoSFBitMask.h
del %COINDIR%\include\Inventor\fields\SoSFBool.h
del %COINDIR%\include\Inventor\fields\SoSFBox2s.h
del %COINDIR%\include\Inventor\fields\SoSFBox2i32.h
del %COINDIR%\include\Inventor\fields\SoSFBox2f.h
del %COINDIR%\include\Inventor\fields\SoSFBox2d.h
del %COINDIR%\include\Inventor\fields\SoSFBox3s.h
del %COINDIR%\include\Inventor\fields\SoSFBox3i32.h
del %COINDIR%\include\Inventor\fields\SoSFBox3f.h
del %COINDIR%\include\Inventor\fields\SoSFBox3d.h
del %COINDIR%\include\Inventor\fields\SoSFColor.h
del %COINDIR%\include\Inventor\fields\SoSFColorRGBA.h
del %COINDIR%\include\Inventor\fields\SoSFDouble.h
del %COINDIR%\include\Inventor\fields\SoSFEngine.h
del %COINDIR%\include\Inventor\fields\SoSFEnum.h
del %COINDIR%\include\Inventor\fields\SoSFFloat.h
del %COINDIR%\include\Inventor\fields\SoSFImage.h
del %COINDIR%\include\Inventor\fields\SoSFImage3.h
del %COINDIR%\include\Inventor\fields\SoSFInt32.h
del %COINDIR%\include\Inventor\fields\SoSFLong.h
del %COINDIR%\include\Inventor\fields\SoSFMatrix.h
del %COINDIR%\include\Inventor\fields\SoSFName.h
del %COINDIR%\include\Inventor\fields\SoSFNode.h
del %COINDIR%\include\Inventor\fields\SoSFPath.h
del %COINDIR%\include\Inventor\fields\SoSFPlane.h
del %COINDIR%\include\Inventor\fields\SoSFRotation.h
del %COINDIR%\include\Inventor\fields\SoSFShort.h
del %COINDIR%\include\Inventor\fields\SoSFString.h
del %COINDIR%\include\Inventor\fields\SoSFTime.h
del %COINDIR%\include\Inventor\fields\SoSFTrigger.h
del %COINDIR%\include\Inventor\fields\SoSFUInt32.h
del %COINDIR%\include\Inventor\fields\SoSFULong.h
del %COINDIR%\include\Inventor\fields\SoSFUShort.h
del %COINDIR%\include\Inventor\fields\SoSFVec2b.h
del %COINDIR%\include\Inventor\fields\SoSFVec2s.h
del %COINDIR%\include\Inventor\fields\SoSFVec2i32.h
del %COINDIR%\include\Inventor\fields\SoSFVec2f.h
del %COINDIR%\include\Inventor\fields\SoSFVec2d.h
del %COINDIR%\include\Inventor\fields\SoSFVec3b.h
del %COINDIR%\include\Inventor\fields\SoSFVec3s.h
del %COINDIR%\include\Inventor\fields\SoSFVec3i32.h
del %COINDIR%\include\Inventor\fields\SoSFVec3f.h
del %COINDIR%\include\Inventor\fields\SoSFVec3d.h
del %COINDIR%\include\Inventor\fields\SoSFVec4b.h
del %COINDIR%\include\Inventor\fields\SoSFVec4ub.h
del %COINDIR%\include\Inventor\fields\SoSFVec4s.h
del %COINDIR%\include\Inventor\fields\SoSFVec4us.h
del %COINDIR%\include\Inventor\fields\SoSFVec4i32.h
del %COINDIR%\include\Inventor\fields\SoSFVec4ui32.h
del %COINDIR%\include\Inventor\fields\SoSFVec4f.h
del %COINDIR%\include\Inventor\fields\SoSFVec4d.h
del %COINDIR%\include\Inventor\fields\SoSField.h
del %COINDIR%\include\Inventor\lists\SbList.h
del %COINDIR%\include\Inventor\lists\SbPList.h
del %COINDIR%\include\Inventor\lists\SbIntList.h
del %COINDIR%\include\Inventor\lists\SbVec3fList.h
del %COINDIR%\include\Inventor\lists\SbStringList.h
del %COINDIR%\include\Inventor\lists\SoActionMethodList.h
del %COINDIR%\include\Inventor\lists\SoAuditorList.h
del %COINDIR%\include\Inventor\lists\SoBaseList.h
del %COINDIR%\include\Inventor\lists\SoCallbackList.h
del %COINDIR%\include\Inventor\lists\SoDetailList.h
del %COINDIR%\include\Inventor\lists\SoEnabledElementsList.h
del %COINDIR%\include\Inventor\lists\SoEngineList.h
del %COINDIR%\include\Inventor\lists\SoEngineOutputList.h
del %COINDIR%\include\Inventor\lists\SoFieldList.h
del %COINDIR%\include\Inventor\lists\SoNodeList.h
del %COINDIR%\include\Inventor\lists\SoPathList.h
del %COINDIR%\include\Inventor\lists\SoPickedPointList.h
del %COINDIR%\include\Inventor\lists\SoTypeList.h
del %COINDIR%\include\Inventor\lock\SoLockMgr.h
del %COINDIR%\include\Inventor\manips\SoClipPlaneManip.h
del %COINDIR%\include\Inventor\manips\SoDirectionalLightManip.h
del %COINDIR%\include\Inventor\manips\SoPointLightManip.h
del %COINDIR%\include\Inventor\manips\SoSpotLightManip.h
del %COINDIR%\include\Inventor\manips\SoTransformManip.h
del %COINDIR%\include\Inventor\manips\SoCenterballManip.h
del %COINDIR%\include\Inventor\manips\SoHandleBoxManip.h
del %COINDIR%\include\Inventor\manips\SoJackManip.h
del %COINDIR%\include\Inventor\manips\SoTabBoxManip.h
del %COINDIR%\include\Inventor\manips\SoTrackballManip.h
del %COINDIR%\include\Inventor\manips\SoTransformBoxManip.h
del %COINDIR%\include\Inventor\manips\SoTransformerManip.h
del %COINDIR%\include\Inventor\misc\CoinResources.h
del %COINDIR%\include\Inventor\misc\SoAuditorList.h
del %COINDIR%\include\Inventor\misc\SoBase.h
del %COINDIR%\include\Inventor\misc\SoBasic.h
del %COINDIR%\include\Inventor\misc\SoByteStream.h
del %COINDIR%\include\Inventor\misc\SoCallbackList.h
del %COINDIR%\include\Inventor\misc\SoChildList.h
del %COINDIR%\include\Inventor\misc\SoContextHandler.h
del %COINDIR%\include\Inventor\misc\SoGLImage.h
del %COINDIR%\include\Inventor\misc\SoGLCubeMapImage.h
del %COINDIR%\include\Inventor\misc\SoGLBigImage.h
del %COINDIR%\include\Inventor\misc\SoNormalGenerator.h
del %COINDIR%\include\Inventor\misc\SoNotification.h
del %COINDIR%\include\Inventor\misc\SoNotRec.h
del %COINDIR%\include\Inventor\misc\SoProto.h
del %COINDIR%\include\Inventor\misc\SoProtoInstance.h
del %COINDIR%\include\Inventor\misc\SoTranReceiver.h
del %COINDIR%\include\Inventor\misc\SoState.h
del %COINDIR%\include\Inventor\misc\SoTranscribe.h
del %COINDIR%\include\Inventor\misc\SoTranSender.h
del %COINDIR%\include\Inventor\misc\SoLightPath.h
del %COINDIR%\include\Inventor\misc\SoTempPath.h
del %COINDIR%\include\Inventor\misc\SoGlyph.h
del %COINDIR%\include\Inventor\misc\SoAudioDevice.h
del %COINDIR%\include\Inventor\misc\SoScriptEngine.h
del %COINDIR%\include\Inventor\misc\SoJavaScriptEngine.h
del %COINDIR%\include\Inventor\misc\SoGLDriverDatabase.h
del %COINDIR%\include\Inventor\navigation\SoScXMLNavigation.h
del %COINDIR%\include\Inventor\navigation\SoScXMLNavigationTarget.h
del %COINDIR%\include\Inventor\navigation\SoScXMLMiscTarget.h
del %COINDIR%\include\Inventor\navigation\SoScXMLPanTarget.h
del %COINDIR%\include\Inventor\navigation\SoScXMLRotateTarget.h
del %COINDIR%\include\Inventor\navigation\SoScXMLSeekTarget.h
del %COINDIR%\include\Inventor\navigation\SoScXMLSpinTarget.h
del %COINDIR%\include\Inventor\navigation\SoScXMLZoomTarget.h
del %COINDIR%\include\Inventor\navigation\SoScXMLDollyTarget.h
del %COINDIR%\include\Inventor\navigation\SoScXMLFlightControlTarget.h
del %COINDIR%\include\Inventor\navigation\SoScXMLMotionTarget.h
del %COINDIR%\include\Inventor\nodekits\SoSubKit.h
del %COINDIR%\include\Inventor\nodekits\SoNodeKit.h
del %COINDIR%\include\Inventor\nodekits\SoNodeKitListPart.h
del %COINDIR%\include\Inventor\nodekits\SoNodekitCatalog.h
del %COINDIR%\include\Inventor\nodekits\SoBaseKit.h
del %COINDIR%\include\Inventor\nodekits\SoAppearanceKit.h
del %COINDIR%\include\Inventor\nodekits\SoCameraKit.h
del %COINDIR%\include\Inventor\nodekits\SoInteractionKit.h
del %COINDIR%\include\Inventor\nodekits\SoLightKit.h
del %COINDIR%\include\Inventor\nodekits\SoSceneKit.h
del %COINDIR%\include\Inventor\nodekits\SoSeparatorKit.h
del %COINDIR%\include\Inventor\nodekits\SoShapeKit.h
del %COINDIR%\include\Inventor\nodekits\SoWrapperKit.h
del %COINDIR%\include\Inventor\nodes\SoAlphaTest.h
del %COINDIR%\include\Inventor\nodes\SoAnnotation.h
del %COINDIR%\include\Inventor\nodes\SoAntiSquish.h
del %COINDIR%\include\Inventor\nodes\SoArray.h
del %COINDIR%\include\Inventor\nodes\SoAsciiText.h
del %COINDIR%\include\Inventor\nodes\SoBaseColor.h
del %COINDIR%\include\Inventor\nodes\SoBlinker.h
del %COINDIR%\include\Inventor\nodes\SoBumpMap.h
del %COINDIR%\include\Inventor\nodes\SoBumpMapCoordinate.h
del %COINDIR%\include\Inventor\nodes\SoBumpMapTransform.h
del %COINDIR%\include\Inventor\nodes\SoCacheHint.h
del %COINDIR%\include\Inventor\nodes\SoCallback.h
del %COINDIR%\include\Inventor\nodes\SoCamera.h
del %COINDIR%\include\Inventor\nodes\SoClipPlane.h
del %COINDIR%\include\Inventor\nodes\SoColorIndex.h
del %COINDIR%\include\Inventor\nodes\SoComplexity.h
del %COINDIR%\include\Inventor\nodes\SoCone.h
del %COINDIR%\include\Inventor\nodes\SoCoordinate3.h
del %COINDIR%\include\Inventor\nodes\SoCoordinate4.h
del %COINDIR%\include\Inventor\nodes\SoCube.h
del %COINDIR%\include\Inventor\nodes\SoCylinder.h
del %COINDIR%\include\Inventor\nodes\SoDepthBuffer.h
del %COINDIR%\include\Inventor\nodes\SoDirectionalLight.h
del %COINDIR%\include\Inventor\nodes\SoDrawStyle.h
del %COINDIR%\include\Inventor\nodes\SoEnvironment.h
del %COINDIR%\include\Inventor\nodes\SoEventCallback.h
del %COINDIR%\include\Inventor\nodes\SoExtSelection.h
del %COINDIR%\include\Inventor\nodes\SoFaceSet.h
del %COINDIR%\include\Inventor\nodes\SoFile.h
del %COINDIR%\include\Inventor\nodes\SoFont.h
del %COINDIR%\include\Inventor\nodes\SoFontStyle.h
del %COINDIR%\include\Inventor\nodes\SoFrustumCamera.h
del %COINDIR%\include\Inventor\nodes\SoGeoOrigin.h
del %COINDIR%\include\Inventor\nodes\SoGeoLocation.h
del %COINDIR%\include\Inventor\nodes\SoGeoSeparator.h
del %COINDIR%\include\Inventor\nodes\SoGeoCoordinate.h
del %COINDIR%\include\Inventor\nodes\SoGroup.h
del %COINDIR%\include\Inventor\nodes\SoGeometryShader.h
del %COINDIR%\include\Inventor\nodes\SoImage.h
del %COINDIR%\include\Inventor\nodes\SoIndexedFaceSet.h
del %COINDIR%\include\Inventor\nodes\SoIndexedLineSet.h
del %COINDIR%\include\Inventor\nodes\SoIndexedMarkerSet.h
del %COINDIR%\include\Inventor\nodes\SoIndexedNurbsCurve.h
del %COINDIR%\include\Inventor\nodes\SoIndexedNurbsSurface.h
del %COINDIR%\include\Inventor\nodes\SoIndexedPointSet.h
del %COINDIR%\include\Inventor\nodes\SoIndexedShape.h
del %COINDIR%\include\Inventor\nodes\SoIndexedTriangleStripSet.h
del %COINDIR%\include\Inventor\nodes\SoInfo.h
del %COINDIR%\include\Inventor\nodes\SoLOD.h
del %COINDIR%\include\Inventor\nodes\SoLabel.h
del %COINDIR%\include\Inventor\nodes\SoLevelOfDetail.h
del %COINDIR%\include\Inventor\nodes\SoLight.h
del %COINDIR%\include\Inventor\nodes\SoLightModel.h
del %COINDIR%\include\Inventor\nodes\SoLineSet.h
del %COINDIR%\include\Inventor\nodes\SoLinearProfile.h
del %COINDIR%\include\Inventor\nodes\SoListener.h
del %COINDIR%\include\Inventor\nodes\SoLocateHighlight.h
del %COINDIR%\include\Inventor\nodes\SoMarkerSet.h
del %COINDIR%\include\Inventor\nodes\SoMaterial.h
del %COINDIR%\include\Inventor\nodes\SoMaterialBinding.h
del %COINDIR%\include\Inventor\nodes\SoMatrixTransform.h
del %COINDIR%\include\Inventor\nodes\SoMultipleCopy.h
del %COINDIR%\include\Inventor\nodes\SoNode.h
del %COINDIR%\include\Inventor\nodes\SoNodes.h
del %COINDIR%\include\Inventor\nodes\SoNonIndexedShape.h
del %COINDIR%\include\Inventor\nodes\SoNormal.h
del %COINDIR%\include\Inventor\nodes\SoNormalBinding.h
del %COINDIR%\include\Inventor\nodes\SoNurbsCurve.h
del %COINDIR%\include\Inventor\nodes\SoNurbsProfile.h
del %COINDIR%\include\Inventor\nodes\SoNurbsSurface.h
del %COINDIR%\include\Inventor\nodes\SoOrthographicCamera.h
del %COINDIR%\include\Inventor\nodes\SoPackedColor.h
del %COINDIR%\include\Inventor\nodes\SoPathSwitch.h
del %COINDIR%\include\Inventor\nodes\SoPendulum.h
del %COINDIR%\include\Inventor\nodes\SoPerspectiveCamera.h
del %COINDIR%\include\Inventor\nodes\SoPickStyle.h
del %COINDIR%\include\Inventor\nodes\SoPointLight.h
del %COINDIR%\include\Inventor\nodes\SoPointSet.h
del %COINDIR%\include\Inventor\nodes\SoPolygonOffset.h
del %COINDIR%\include\Inventor\nodes\SoProfile.h
del %COINDIR%\include\Inventor\nodes\SoProfileCoordinate2.h
del %COINDIR%\include\Inventor\nodes\SoProfileCoordinate3.h
del %COINDIR%\include\Inventor\nodes\SoQuadMesh.h
del %COINDIR%\include\Inventor\nodes\SoResetTransform.h
del %COINDIR%\include\Inventor\nodes\SoReversePerspectiveCamera.h
del %COINDIR%\include\Inventor\nodes\SoRotation.h
del %COINDIR%\include\Inventor\nodes\SoRotationXYZ.h
del %COINDIR%\include\Inventor\nodes\SoRotor.h
del %COINDIR%\include\Inventor\nodes\SoScale.h
del %COINDIR%\include\Inventor\nodes\SoSceneTexture2.h
del %COINDIR%\include\Inventor\nodes\SoSceneTextureCubeMap.h
del %COINDIR%\include\Inventor\nodes\SoSelection.h
del %COINDIR%\include\Inventor\nodes\SoSeparator.h
del %COINDIR%\include\Inventor\nodes\SoShape.h
del %COINDIR%\include\Inventor\nodes\SoShapeHints.h
del %COINDIR%\include\Inventor\nodes\SoShuttle.h
del %COINDIR%\include\Inventor\nodes\SoSphere.h
del %COINDIR%\include\Inventor\nodes\SoSpotLight.h
del %COINDIR%\include\Inventor\nodes\SoSubNode.h
del %COINDIR%\include\Inventor\nodes\SoSurroundScale.h
del %COINDIR%\include\Inventor\nodes\SoSwitch.h
del %COINDIR%\include\Inventor\nodes\SoText2.h
del %COINDIR%\include\Inventor\nodes\SoText3.h
del %COINDIR%\include\Inventor\nodes\SoTexture.h
del %COINDIR%\include\Inventor\nodes\SoTexture2.h
del %COINDIR%\include\Inventor\nodes\SoTexture2Transform.h
del %COINDIR%\include\Inventor\nodes\SoTexture3.h
del %COINDIR%\include\Inventor\nodes\SoTexture3Transform.h
del %COINDIR%\include\Inventor\nodes\SoTextureCombine.h
del %COINDIR%\include\Inventor\nodes\SoTextureCoordinate2.h
del %COINDIR%\include\Inventor\nodes\SoTextureCoordinate3.h
del %COINDIR%\include\Inventor\nodes\SoTextureCoordinateBinding.h
del %COINDIR%\include\Inventor\nodes\SoTextureCoordinateCube.h
del %COINDIR%\include\Inventor\nodes\SoTextureCoordinateCylinder.h
del %COINDIR%\include\Inventor\nodes\SoTextureCoordinateSphere.h
del %COINDIR%\include\Inventor\nodes\SoTextureCoordinateDefault.h
del %COINDIR%\include\Inventor\nodes\SoTextureCoordinateEnvironment.h
del %COINDIR%\include\Inventor\nodes\SoTextureCoordinateFunction.h
del %COINDIR%\include\Inventor\nodes\SoTextureCoordinatePlane.h
del %COINDIR%\include\Inventor\nodes\SoTextureCoordinateNormalMap.h
del %COINDIR%\include\Inventor\nodes\SoTextureCoordinateReflectionMap.h
del %COINDIR%\include\Inventor\nodes\SoTextureCoordinateObject.h
del %COINDIR%\include\Inventor\nodes\SoTextureCubeMap.h
del %COINDIR%\include\Inventor\nodes\SoTextureMatrixTransform.h
del %COINDIR%\include\Inventor\nodes\SoTextureScalePolicy.h
del %COINDIR%\include\Inventor\nodes\SoTextureUnit.h
del %COINDIR%\include\Inventor\nodes\SoTransform.h
del %COINDIR%\include\Inventor\nodes\SoTransformSeparator.h
del %COINDIR%\include\Inventor\nodes\SoTransformation.h
del %COINDIR%\include\Inventor\nodes\SoTranslation.h
del %COINDIR%\include\Inventor\nodes\SoTransparencyType.h
del %COINDIR%\include\Inventor\nodes\SoTriangleStripSet.h
del %COINDIR%\include\Inventor\nodes\SoUnits.h
del %COINDIR%\include\Inventor\nodes\SoVertexProperty.h
del %COINDIR%\include\Inventor\nodes\SoVertexAttribute.h
del %COINDIR%\include\Inventor\nodes\SoVertexAttributeBinding.h
del %COINDIR%\include\Inventor\nodes\SoVertexShape.h
del %COINDIR%\include\Inventor\nodes\SoWWWAnchor.h
del %COINDIR%\include\Inventor\nodes\SoWWWInline.h
del %COINDIR%\include\Inventor\nodes\SoFragmentShader.h
del %COINDIR%\include\Inventor\nodes\SoShaderObject.h
del %COINDIR%\include\Inventor\nodes\SoShaderParameter.h
del %COINDIR%\include\Inventor\nodes\SoShaderProgram.h
del %COINDIR%\include\Inventor\nodes\SoVertexShader.h
del %COINDIR%\include\Inventor\projectors\SbProjectors.h
del %COINDIR%\include\Inventor\projectors\SbCylinderPlaneProjector.h
del %COINDIR%\include\Inventor\projectors\SbCylinderProjector.h
del %COINDIR%\include\Inventor\projectors\SbCylinderSectionProjector.h
del %COINDIR%\include\Inventor\projectors\SbCylinderSheetProjector.h
del %COINDIR%\include\Inventor\projectors\SbLineProjector.h
del %COINDIR%\include\Inventor\projectors\SbPlaneProjector.h
del %COINDIR%\include\Inventor\projectors\SbProjector.h
del %COINDIR%\include\Inventor\projectors\SbSpherePlaneProjector.h
del %COINDIR%\include\Inventor\projectors\SbSphereProjector.h
del %COINDIR%\include\Inventor\projectors\SbSphereSectionProjector.h
del %COINDIR%\include\Inventor\projectors\SbSphereSheetProjector.h
del %COINDIR%\include\Inventor\sensors\SoSensors.h
del %COINDIR%\include\Inventor\sensors\SoAlarmSensor.h
del %COINDIR%\include\Inventor\sensors\SoDataSensor.h
del %COINDIR%\include\Inventor\sensors\SoDelayQueueSensor.h
del %COINDIR%\include\Inventor\sensors\SoFieldSensor.h
del %COINDIR%\include\Inventor\sensors\SoIdleSensor.h
del %COINDIR%\include\Inventor\sensors\SoNodeSensor.h
del %COINDIR%\include\Inventor\sensors\SoOneShotSensor.h
del %COINDIR%\include\Inventor\sensors\SoPathSensor.h
del %COINDIR%\include\Inventor\sensors\SoSensor.h
del %COINDIR%\include\Inventor\sensors\SoSensorManager.h
del %COINDIR%\include\Inventor\sensors\SoTimerQueueSensor.h
del %COINDIR%\include\Inventor\sensors\SoTimerSensor.h
del %COINDIR%\include\Inventor\system\gl.h
del %COINDIR%\include\Inventor\system\gl-headers.h
del %COINDIR%\include\Inventor\system\inttypes.h
del %COINDIR%\include\Inventor\threads\SbThread.h
del %COINDIR%\include\Inventor\threads\SbMutex.h
del %COINDIR%\include\Inventor\threads\SbThreadMutex.h
del %COINDIR%\include\Inventor\threads\SbRWMutex.h
del %COINDIR%\include\Inventor\threads\SbCondVar.h
del %COINDIR%\include\Inventor\threads\SbStorage.h
del %COINDIR%\include\Inventor\threads\SbTypedStorage.h
del %COINDIR%\include\Inventor\threads\SbFifo.h
del %COINDIR%\include\Inventor\threads\SbBarrier.h
del %COINDIR%\include\Inventor\threads\SbThreadAutoLock.h
del %COINDIR%\include\Inventor\tools\SbPimplPtr.h
del %COINDIR%\include\Inventor\tools\SbPimplPtr.hpp
del %COINDIR%\include\Inventor\tools\SbLazyPimplPtr.h
del %COINDIR%\include\Inventor\tools\SbLazyPimplPtr.hpp
del %COINDIR%\include\Inventor\scxml\ScXML.h
del %COINDIR%\include\Inventor\scxml\ScXMLSubObject.h
del %COINDIR%\include\Inventor\scxml\ScXMLObject.h
del %COINDIR%\include\Inventor\scxml\ScXMLEvent.h
del %COINDIR%\include\Inventor\scxml\ScXMLEventTarget.h
del %COINDIR%\include\Inventor\scxml\ScXMLStateMachine.h
del %COINDIR%\include\Inventor\scxml\ScXMLDocument.h
del %COINDIR%\include\Inventor\scxml\ScXMLElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLAbstractStateElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLExecutableElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLAnchorElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLAssignElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLContentElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLDataElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLDataModelElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLElseElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLElseIfElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLEventElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLFinalElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLFinalizeElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLHistoryElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLIfElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLInitialElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLInvokeElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLLogElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLOnEntryElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLOnExitElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLParallelElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLParamElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLScriptElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLScxmlElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLSendElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLStateElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLTransitionElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLValidateElt.h
del %COINDIR%\include\Inventor\scxml\ScXMLEvaluator.h
del %COINDIR%\include\Inventor\scxml\ScXMLMinimumEvaluator.h
del %COINDIR%\include\Inventor\scxml\ScXMLXPathEvaluator.h
del %COINDIR%\include\Inventor\scxml\ScXMLECMAScriptEvaluator.h
del %COINDIR%\include\Inventor\scxml\ScXMLCoinEvaluator.h
del %COINDIR%\include\Inventor\scxml\SoScXMLEvent.h
del %COINDIR%\include\Inventor\scxml\SoScXMLStateMachine.h
del %COINDIR%\include\Inventor\Sb.h
del %COINDIR%\include\Inventor\SbBSPTree.h
del %COINDIR%\include\Inventor\SbBasic.h
del %COINDIR%\include\Inventor\SbBox.h
del %COINDIR%\include\Inventor\SbBox2s.h
del %COINDIR%\include\Inventor\SbBox2i32.h
del %COINDIR%\include\Inventor\SbBox2f.h
del %COINDIR%\include\Inventor\SbBox2d.h
del %COINDIR%\include\Inventor\SbBox3d.h
del %COINDIR%\include\Inventor\SbBox3f.h
del %COINDIR%\include\Inventor\SbBox3i32.h
del %COINDIR%\include\Inventor\SbBox3s.h
del %COINDIR%\include\Inventor\SbByteBuffer.h
del %COINDIR%\include\Inventor\SbByteBuffer.h
del %COINDIR%\include\Inventor\SbClip.h
del %COINDIR%\include\Inventor\SbColor.h
del %COINDIR%\include\Inventor\SbColor4f.h
del %COINDIR%\include\Inventor\SbColorRGBA.h
del %COINDIR%\include\Inventor\SbCylinder.h
del %COINDIR%\include\Inventor\SbDPLine.h
del %COINDIR%\include\Inventor\SbDPLinear.h
del %COINDIR%\include\Inventor\SbDPMatrix.h
del %COINDIR%\include\Inventor\SbDPPlane.h
del %COINDIR%\include\Inventor\SbDPRotation.h
del %COINDIR%\include\Inventor\SbDPViewVolume.h
del %COINDIR%\include\Inventor\SbDict.h
del %COINDIR%\include\Inventor\SbHeap.h
del %COINDIR%\include\Inventor\SbImage.h
del %COINDIR%\include\Inventor\SbLine.h
del %COINDIR%\include\Inventor\SbLinear.h
del %COINDIR%\include\Inventor\SbMatrix.h
del %COINDIR%\include\Inventor\SbName.h
del %COINDIR%\include\Inventor\SbOctTree.h
del %COINDIR%\include\Inventor\SbPList.h
del %COINDIR%\include\Inventor\SbPlane.h
del %COINDIR%\include\Inventor\SbRotation.h
del %COINDIR%\include\Inventor\SbSphere.h
del %COINDIR%\include\Inventor\SbString.h
del %COINDIR%\include\Inventor\SbTesselator.h
del %COINDIR%\include\Inventor\SbTime.h
del %COINDIR%\include\Inventor\SbTypeInfo.h
del %COINDIR%\include\Inventor\SbVec.h
del %COINDIR%\include\Inventor\SbVec2b.h
del %COINDIR%\include\Inventor\SbVec2ub.h
del %COINDIR%\include\Inventor\SbVec2s.h
del %COINDIR%\include\Inventor\SbVec2us.h
del %COINDIR%\include\Inventor\SbVec2i32.h
del %COINDIR%\include\Inventor\SbVec2ui32.h
del %COINDIR%\include\Inventor\SbVec2f.h
del %COINDIR%\include\Inventor\SbVec2d.h
del %COINDIR%\include\Inventor\SbVec3b.h
del %COINDIR%\include\Inventor\SbVec3ub.h
del %COINDIR%\include\Inventor\SbVec3s.h
del %COINDIR%\include\Inventor\SbVec3us.h
del %COINDIR%\include\Inventor\SbVec3i32.h
del %COINDIR%\include\Inventor\SbVec3ui32.h
del %COINDIR%\include\Inventor\SbVec3f.h
del %COINDIR%\include\Inventor\SbVec3d.h
del %COINDIR%\include\Inventor\SbVec4b.h
del %COINDIR%\include\Inventor\SbVec4ub.h
del %COINDIR%\include\Inventor\SbVec4s.h
del %COINDIR%\include\Inventor\SbVec4us.h
del %COINDIR%\include\Inventor\SbVec4i32.h
del %COINDIR%\include\Inventor\SbVec4ui32.h
del %COINDIR%\include\Inventor\SbVec4f.h
del %COINDIR%\include\Inventor\SbVec4d.h
del %COINDIR%\include\Inventor\SbViewVolume.h
del %COINDIR%\include\Inventor\SbViewportRegion.h
del %COINDIR%\include\Inventor\SbXfBox3f.h
del %COINDIR%\include\Inventor\SbXfBox3d.h
del %COINDIR%\include\Inventor\So.h
del %COINDIR%\include\Inventor\SoDB.h
del %COINDIR%\include\Inventor\SoFullPath.h
del %COINDIR%\include\Inventor\SoInput.h
del %COINDIR%\include\Inventor\SoInteraction.h
del %COINDIR%\include\Inventor\SoLists.h
del %COINDIR%\include\Inventor\SoNodeKitPath.h
del %COINDIR%\include\Inventor\SoOffscreenRenderer.h
del %COINDIR%\include\Inventor\SoOutput.h
del %COINDIR%\include\Inventor\SoPath.h
del %COINDIR%\include\Inventor\SoPickedPoint.h
del %COINDIR%\include\Inventor\SoPrimitiveVertex.h
del %COINDIR%\include\Inventor\SoSceneManager.h
del %COINDIR%\include\Inventor\SoRenderManager.h
del %COINDIR%\include\Inventor\SoEventManager.h
del %COINDIR%\include\Inventor\SoType.h
del %COINDIR%\include\Inventor\non_winsys.h
del %COINDIR%\include\Inventor\oivwin32.h
del %COINDIR%\include\SoWinEnterScope.h
del %COINDIR%\include\SoWinLeaveScope.h
del %COINDIR%\include\SoDebug.h
