#ifndef COIN_QTWIDGET_H
#define COIN_QTWIDGET_H

/**************************************************************************\
 * Copyright (c) Kongsberg Oil & Gas Technologies AS
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
\**************************************************************************/

#include <qgl.h>

#include <Inventor/SbBasic.h>

class SoNode;
class QDialog;

// *************************************************************************

//FIXME: Split this class in one CoinQt3Widget and one CoinQt4Widget
//file, and port the last one to Qt4. BFG 20090316

class CoinQtWidget : public QGLWidget {
  typedef QGLWidget inherited;
  Q_OBJECT

public:
  CoinQtWidget(QWidget * parent = NULL, const char * name = NULL,
               const QGLWidget * shareWidget = NULL, WFlags f = 0);
  CoinQtWidget(const QGLFormat & format, QWidget * parent = NULL,
               const char * name = NULL, const QGLWidget * shareWidget = NULL,
               WFlags f = 0);
  virtual ~CoinQtWidget(void);

  virtual SbBool setSceneGraph(SoNode * root, SbBool managed = FALSE);
  virtual SoNode * getSceneGraph(void) const;

public slots:
  virtual void tick(void);
  virtual void about(void);

protected:
  virtual void resizeGL(int width, int height);
  virtual void initializeGL(void);
  virtual void paintGL(void);
  virtual void glInit(void);
  virtual void glDraw(void);

  virtual void mousePressEvent(QMouseEvent * event);
  virtual void mouseReleaseEvent(QMouseEvent * event);
  virtual void mouseMoveEvent(QMouseEvent * event);
  virtual void keyPressEvent(QKeyEvent * event);
  virtual void keyReleaseEvent(QKeyEvent * event);

  virtual void updateGL(void);

private:
  class CoinQtWidgetP * internals;

}; // CoinQtWidget

// *************************************************************************

#endif /* !COIN_QTWIDGET_H */
