#!/bin/sh

./extractfont 7 13 0 1 8 5 16 font13.tif >font13.ic

./extractfont 10 17 0 1 12 6 16 font17.tif >font17.ic

./extractfont 14 25 0 2 17 11 16 font25.tif >font25.ic

./extractfont 19 33 0 3 25 16 16 font33.tif >font33.ic
