/**************************************************************************\
 * Copyright (c) Kongsberg Oil & Gas Technologies AS
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
\**************************************************************************/

#include "SoAccumulatedElement.cpp"
#include "SoAmbientColorElement.cpp"
#include "SoAnnoText3CharOrientElement.cpp"
#include "SoAnnoText3FontSizeHintElement.cpp"
#include "SoAnnoText3RenderPrintElement.cpp"
#include "SoBBoxModelMatrixElement.cpp"
#include "SoBumpMapElement.cpp"
#include "SoBumpMapCoordinateElement.cpp"
#include "SoBumpMapMatrixElement.cpp"
#include "SoBumpMappingPropertyElement.cpp"
#include "SoCacheElement.cpp"
#include "SoCacheHintElement.cpp"
#include "SoClipPlaneElement.cpp"
#include "SoComplexityElement.cpp"
#include "SoComplexityTypeElement.cpp"
#include "SoCoordinateElement.cpp"
#include "SoCreaseAngleElement.cpp"
#include "SoCullElement.cpp"
#include "SoDecimationPercentageElement.cpp"
#include "SoDecimationTypeElement.cpp"
#include "SoDepthBufferElement.cpp"
#include "SoDiffuseColorElement.cpp"
#include "SoDrawStyleElement.cpp"
#include "SoElement.cpp"
#include "SoEmissiveColorElement.cpp"
#include "SoEnvironmentElement.cpp"
#include "SoFloatElement.cpp"
#include "SoFocalDistanceElement.cpp"
#include "SoFontNameElement.cpp"
#include "SoFontSizeElement.cpp"
#include "SoInt32Element.cpp"
#include "SoLazyElement.cpp"
#include "SoLightAttenuationElement.cpp"
#include "SoLightElement.cpp"
#include "SoLightModelElement.cpp"
#include "SoLinePatternElement.cpp"
#include "SoLineWidthElement.cpp"
#include "SoListenerDopplerElement.cpp"
#include "SoListenerGainElement.cpp"
#include "SoListenerOrientationElement.cpp"
#include "SoListenerPositionElement.cpp"
#include "SoLocalBBoxMatrixElement.cpp"
#include "SoMaterialBindingElement.cpp"
#include "SoModelMatrixElement.cpp"
#include "SoMultiTextureCoordinateElement.cpp"
#include "SoMultiTextureEnabledElement.cpp"
#include "SoMultiTextureImageElement.cpp"
#include "SoMultiTextureMatrixElement.cpp"
#include "SoNormalBindingElement.cpp"
#include "SoNormalElement.cpp"
#include "SoOverrideElement.cpp"
#include "SoPickRayElement.cpp"
#include "SoPickStyleElement.cpp"
#include "SoPointSizeElement.cpp"
#include "SoPolygonOffsetElement.cpp"
#include "SoProfileCoordinateElement.cpp"
#include "SoProfileElement.cpp"
#include "SoProjectionMatrixElement.cpp"
#include "SoReplacedElement.cpp"
#include "SoShapeHintsElement.cpp"
#include "SoShapeStyleElement.cpp"
#include "SoShininessElement.cpp"
#include "SoSoundElement.cpp"
#include "SoSpecularColorElement.cpp"
#include "SoSwitchElement.cpp"
#include "SoTextOutlineEnabledElement.cpp"
#include "SoTextureCombineElement.cpp"
#include "SoTextureCoordinateBindingElement.cpp"
#include "SoTextureOverrideElement.cpp"
#include "SoTextureQualityElement.cpp"
#include "SoTextureScalePolicyElement.cpp"
#include "SoTextureScaleQualityElement.cpp"
#include "SoTextureUnitElement.cpp"
#include "SoTransparencyElement.cpp"
#include "SoUnitsElement.cpp"
#include "SoVertexAttributeBindingElement.cpp"
#include "SoVertexAttributeElement.cpp"
#include "SoViewVolumeElement.cpp"
#include "SoViewingMatrixElement.cpp"
#include "SoViewportRegionElement.cpp"
#include "SoWindowElement.cpp"
