#ifndef COIN_VERTEXARRAYINDEXER_H
#define COIN_VERTEXARRAYINDEXER_H

/**************************************************************************\
 * Copyright (c) Kongsberg Oil & Gas Technologies AS
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
\**************************************************************************/

#ifndef COIN_INTERNAL
#error this is a private header file
#endif /* !COIN_INTERNAL */

#include <Inventor/lists/SbList.h>
#include <Inventor/system/gl.h>
#include <Inventor/C/glue/gl.h>
#include <stdlib.h>

class SoVBO;

class SoVertexArrayIndexer {
public:
  SoVertexArrayIndexer(void);
  ~SoVertexArrayIndexer();

  void addTriangle(const int32_t v0,
                   const int32_t v1,
                   const int32_t v2);
  void addLine(const int32_t v0,
               const int32_t v1);
  void addPoint(const int32_t v0);


  void addQuad(const int32_t v0,
               const int32_t v1,
               const int32_t v2,
               const int32_t v3);

  void beginTarget(GLenum target);
  void targetVertex(GLenum target, const int32_t v);
  void endTarget(GLenum target);

  void close(void);
  void render(const cc_glglue * glue, const SbBool renderasvbo, const uint32_t vbocontextid);

  int getNumVertices(void);
  int getNumIndices(void) const;
  const GLint * getIndices(void) const;
  GLint * getWriteableIndices(void);

private:
  void addIndex(int32_t i);
  void sort_triangles(void);
  void sort_lines(void);
  SoVertexArrayIndexer * getNext(void);

  GLenum target;
  SoVertexArrayIndexer * next;

  int targetcounter;
  SbList <GLsizei> countarray;
  SbList <const GLint *> ciarray;
  SbList <GLint> indexarray;
  SoVBO * vbo;
  SbBool use_shorts;
};

#endif // COIN_VERTEXARRAYINDEXER_H
