/**************************************************************************\
 * Copyright (c) Kongsberg Oil & Gas Technologies AS
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
\**************************************************************************/

#include "SbBox2s.cpp"
#include "SbBox2i32.cpp"
#include "SbBox2f.cpp"
#include "SbBox2d.cpp"
#include "SbBox3s.cpp"
#include "SbBox3i32.cpp"
#include "SbBox3f.cpp"
#include "SbBox3d.cpp"
#include "SbXfBox3f.cpp"
#include "SbXfBox3d.cpp"

#include "SbBSPTree.cpp"
#include "SbClip.cpp"
#include "SbColor.cpp"
#include "SbColor4f.cpp"
#include "SbCylinder.cpp"
#include "SbDict.cpp"
#include "SbHeap.cpp"
#include "SbImage.cpp"
#include "SbLine.cpp"
#include "SbDPLine.cpp"
#include "SbMatrix.cpp"
#include "SbDPMatrix.cpp"
#include "SbName.cpp"
#include "SbOctTree.cpp"
#include "SbPlane.cpp"
#include "SbDPPlane.cpp"
#include "SbRotation.cpp"
#include "SbDPRotation.cpp"
#include "SbSphere.cpp"
#include "SbString.cpp"
#include "SbTesselator.cpp"
#include "SbGLUTessellator.cpp"
#include "SbTime.cpp"
#include "SbByteBuffer.cpp"

#include "SbVec2b.cpp"
#include "SbVec2ub.cpp"
#include "SbVec2s.cpp"
#include "SbVec2us.cpp"
#include "SbVec2i32.cpp"
#include "SbVec2ui32.cpp"
#include "SbVec2f.cpp"
#include "SbVec2d.cpp"
#include "SbVec3b.cpp"
#include "SbVec3ub.cpp"
#include "SbVec3s.cpp"
#include "SbVec3us.cpp"
#include "SbVec3i32.cpp"
#include "SbVec3ui32.cpp"
#include "SbVec3f.cpp"
#include "SbVec3d.cpp"
#include "SbVec4b.cpp"
#include "SbVec4ub.cpp"
#include "SbVec4s.cpp"
#include "SbVec4us.cpp"
#include "SbVec4i32.cpp"
#include "SbVec4ui32.cpp"
#include "SbVec4f.cpp"
#include "SbVec4d.cpp"

#include "SbViewVolume.cpp"
#include "SbDPViewVolume.cpp"
#include "SbViewportRegion.cpp"

#include "dict.cpp"
#include "dynarray.cpp"
#include "hash.cpp"
#include "heap.cpp"
#include "list.cpp"
#include "memalloc.cpp"
#include "namemap.cpp"
#include "rbptree.cpp"
#include "string.cpp"
#include "time.cpp"
