#include "SoGeoCoordinate.cpp"
#include "SoGeoElement.cpp"
#include "SoGeoLocation.cpp"
#include "SoGeoOrigin.cpp"
#include "SoGeoSeparator.cpp"
#include "SbGeoAngle.cpp"
#include "SbGeoEllipsoid.cpp"
#include "SbGeoProjection.cpp"
#include "SbPolarStereographic.cpp"
#include "SbUTMProjection.cpp"
#include "SoGeo.cpp"
