/**************************************************************************\
 * Copyright (c) Kongsberg Oil & Gas Technologies AS
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
\**************************************************************************/

#include "SoAlphaTest.cpp"
#include "SoAnnotation.cpp"
#include "SoAntiSquish.cpp"
#include "SoArray.cpp"
#include "SoBaseColor.cpp"
#include "SoBlinker.cpp"
#include "SoBumpMap.cpp"
#include "SoBumpMapCoordinate.cpp"
#include "SoBumpMapTransform.cpp"
#include "SoCacheHint.cpp"
#include "SoCallback.cpp"
#include "SoCamera.cpp"
#include "SoClipPlane.cpp"
#include "SoColorIndex.cpp"
#include "SoComplexity.cpp"
#include "SoCoordinate3.cpp"
#include "SoCoordinate4.cpp"
#include "SoDepthBuffer.cpp"
#include "SoDirectionalLight.cpp"
#include "SoDrawStyle.cpp"
#include "SoEnvironment.cpp"
#include "SoEventCallback.cpp"
#include "SoExtSelection.cpp"
#include "SoFile.cpp"
#include "SoFont.cpp"
#include "SoFontStyle.cpp"
#include "SoFrustumCamera.cpp"
#include "SoGroup.cpp"
#include "SoInfo.cpp"
#include "SoLOD.cpp"
#include "SoLabel.cpp"
#include "SoLevelOfDetail.cpp"
#include "SoLight.cpp"
#include "SoLightModel.cpp"
#include "SoLinearProfile.cpp"
#include "SoListener.cpp"
#include "SoLocateHighlight.cpp"
#include "SoMaterial.cpp"
#include "SoMaterialBinding.cpp"
#include "SoMatrixTransform.cpp"
#include "SoMultipleCopy.cpp"
#include "SoNode.cpp"
#include "SoNormal.cpp"
#include "SoNormalBinding.cpp"
#include "SoNurbsProfile.cpp"
#include "SoOrthographicCamera.cpp"
#include "SoPackedColor.cpp"
#include "SoPathSwitch.cpp"
#include "SoPendulum.cpp"
#include "SoPerspectiveCamera.cpp"
#include "SoReversePerspectiveCamera.cpp"
#include "SoPickStyle.cpp"
#include "SoPointLight.cpp"
#include "SoPolygonOffset.cpp"
#include "SoProfile.cpp"
#include "SoProfileCoordinate2.cpp"
#include "SoProfileCoordinate3.cpp"
#include "SoResetTransform.cpp"
#include "SoRotation.cpp"
#include "SoRotationXYZ.cpp"
#include "SoRotor.cpp"
#include "SoScale.cpp"
#include "SoSceneTexture2.cpp"
#include "SoSceneTextureCubeMap.cpp"
#include "SoSelection.cpp"
#include "SoSeparator.cpp"
#include "SoShapeHints.cpp"
#include "SoShuttle.cpp"
#include "SoSpotLight.cpp"
#include "SoSurroundScale.cpp"
#include "SoSwitch.cpp"
#include "SoTexture.cpp"
#include "SoTexture2.cpp"
#include "SoTexture2Transform.cpp"
#include "SoTexture3.cpp"
#include "SoTexture3Transform.cpp"
#include "SoTextureCombine.cpp"
#include "SoTextureCoordinate2.cpp"
#include "SoTextureCoordinate3.cpp"
#include "SoTextureCoordinateBinding.cpp"
#include "SoTextureCoordinateCube.cpp"
#include "SoTextureCoordinateCylinder.cpp"
#include "SoTextureCoordinateSphere.cpp"
#include "SoTextureCoordinateDefault.cpp"
#include "SoTextureCoordinateEnvironment.cpp"
#include "SoTextureCoordinateFunction.cpp"
#include "SoTextureCoordinatePlane.cpp"
#include "SoTextureCoordinateNormalMap.cpp"
#include "SoTextureCoordinateReflectionMap.cpp"
#include "SoTextureCoordinateObject.cpp"
#include "SoTextureCubeMap.cpp"
#include "SoTextureMatrixTransform.cpp"
#include "SoTextureScalePolicy.cpp"
#include "SoTextureUnit.cpp"
#include "SoTransform.cpp"
#include "SoTransformSeparator.cpp"
#include "SoTransformation.cpp"
#include "SoTranslation.cpp"
#include "SoTransparencyType.cpp"
#include "SoUnits.cpp"
#include "SoUnknownNode.cpp"
#include "SoVertexProperty.cpp"
#include "SoWWWAnchor.cpp"
#include "SoWWWInline.cpp"
#include "SoVertexAttribute.cpp"
#include "SoVertexAttributeBinding.cpp"
