#ifndef COIN_GLUE_BZIP2_H
#define COIN_GLUE_BZIP2_H

/**************************************************************************\
 * Copyright (c) Kongsberg Oil & Gas Technologies AS
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
\**************************************************************************/

#ifndef COIN_INTERNAL
#error this is a private header file
#endif

#include <Inventor/C/basic.h>
#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#if 0 /* to get proper auto-indentation in emacs */
}
#endif /* emacs indentation */

int cc_bzglue_available(void);

const char * cc_bzglue_BZ2_bzlibVersion(void);
void * cc_bzglue_BZ2_bzReadOpen(int * bzerror,
                                FILE * f,
                                int verbosity,
                                int bzsmall,
                                void * unused,
                                int nunused);
void cc_bzglue_BZ2_bzReadClose(int * bzerror, 
                               void * bzfile); 
int cc_bzglue_BZ2_bzRead(int * bzerror, 
                         void * bzfile, 
                         void * buf, 
                         int len);
void * cc_bzglue_BZ2_bzWriteOpen(int * bzerror,      
                                 FILE * fp, 
                                 int blocksize100k, 
                                 int verbosity, 
                                 int workfactor); 
void cc_bzglue_BZ2_bzWriteClose(int * bzerror, 
                                void * bzfile, 
                                int abandon, 
                                unsigned int * nbytesin, 
                                unsigned int * nbytesout);
void cc_bzglue_BZ2_bzWrite(int * bzerror, 
                           void * bzfile, 
                           void * buf, 
                           int len);
#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* !COIN_GLUE_BZIP2_H */
