/**************************************************************************\
 * Copyright (c) Kongsberg Oil & Gas Technologies AS
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
\**************************************************************************/

#include "SoField.cpp"
#include "SoFieldContainer.cpp"
#include "SoFieldData.cpp"
#include "SoMFBitMask.cpp"
#include "SoMFBool.cpp"
#include "SoMFColor.cpp"
#include "SoMFColorRGBA.cpp"
#include "SoMFDouble.cpp"
#include "SoMFEngine.cpp"
#include "SoMFEnum.cpp"
#include "SoMFFloat.cpp"
#include "SoMFInt32.cpp"
#include "SoMFMatrix.cpp"
#include "SoMFName.cpp"
#include "SoMFNode.cpp"
#include "SoMFPath.cpp"
#include "SoMFPlane.cpp"
#include "SoMFRotation.cpp"
#include "SoMFShort.cpp"
#include "SoMFString.cpp"
#include "SoMFTime.cpp"
#include "SoMFUInt32.cpp"
#include "SoMFUShort.cpp"
#include "SoMFVec2b.cpp"
#include "SoMFVec2s.cpp"
#include "SoMFVec2i32.cpp"
#include "SoMFVec2f.cpp"
#include "SoMFVec2d.cpp"
#include "SoMFVec3b.cpp"
#include "SoMFVec3s.cpp"
#include "SoMFVec3i32.cpp"
#include "SoMFVec3f.cpp"
#include "SoMFVec3d.cpp"
#include "SoMFVec4b.cpp"
#include "SoMFVec4ub.cpp"
#include "SoMFVec4s.cpp"
#include "SoMFVec4us.cpp"
#include "SoMFVec4i32.cpp"
#include "SoMFVec4ui32.cpp"
#include "SoMFVec4f.cpp"
#include "SoMFVec4d.cpp"
#include "SoMField.cpp"
#include "SoSFBitMask.cpp"
#include "SoSFBool.cpp"
#include "SoSFColor.cpp"
#include "SoSFColorRGBA.cpp"
#include "SoSFDouble.cpp"
#include "SoSFEngine.cpp"
#include "SoSFEnum.cpp"
#include "SoSFFloat.cpp"
#include "SoSFImage.cpp"
#include "SoSFImage3.cpp"
#include "SoSFInt32.cpp"
#include "SoSFMatrix.cpp"
#include "SoSFName.cpp"
#include "SoSFNode.cpp"
#include "SoSFPath.cpp"
#include "SoSFPlane.cpp"
#include "SoSFRotation.cpp"
#include "SoSFShort.cpp"
#include "SoSFString.cpp"
#include "SoSFTime.cpp"
#include "SoSFTrigger.cpp"
#include "SoSFUInt32.cpp"
#include "SoSFUShort.cpp"
#include "SoSFVec2b.cpp"
#include "SoSFVec2s.cpp"
#include "SoSFVec2i32.cpp"
#include "SoSFVec2f.cpp"
#include "SoSFVec2d.cpp"
#include "SoSFVec3b.cpp"
#include "SoSFVec3s.cpp"
#include "SoSFVec3i32.cpp"
#include "SoSFVec3f.cpp"
#include "SoSFVec3d.cpp"
#include "SoSFVec4b.cpp"
#include "SoSFVec4ub.cpp"
#include "SoSFVec4s.cpp"
#include "SoSFVec4us.cpp"
#include "SoSFVec4i32.cpp"
#include "SoSFVec4ui32.cpp"
#include "SoSFVec4f.cpp"
#include "SoSFVec4d.cpp"
#include "SoSFBox2s.cpp"
#include "SoSFBox2i32.cpp"
#include "SoSFBox2f.cpp"
#include "SoSFBox2d.cpp"
#include "SoSFBox3s.cpp"
#include "SoSFBox3i32.cpp"
#include "SoSFBox3f.cpp"
#include "SoSFBox3d.cpp"
#include "SoSField.cpp"
#include "SoGlobalField.cpp"
#include "shared.cpp"
